/* PAA - Steineruv problem */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>

#define MAXCOORDS 200
#define TESTNAMELEN 80
#define LINELEN 256

#define SCALE 4
#define ALLOWEDTIME 60

#define MAXSTPTS  1000  /* max number of Steiner points for the proper
                           best-only search to be completed in time */

#define VIEW 0    /* graphical output */
#define INTBITS 32 /* bits in an unsigned integer */

#if VIEW           /* if VIEW is set, use VGAlib to display */
#include <vga.h>
#endif

#define INTBITS 32            /* bits in an unsigned integer */

char test_name[TESTNAMELEN] ; /* test instance name */ 

FILE *texf ;                  /* output TeX file stream */ 
 
struct point { int x,y ; } coords[MAXCOORDS] ; /* input coordinates */
int numpoints ;                                /* number of points */

volatile sig_atomic_t keep_going ;             /* a flag to stop on timeout */


/* ---------------------------------------------------------------------- */

/* read input and store it */

int read_input()      
{ char buf[LINELEN] ;

  fgets(test_name,TESTNAMELEN,stdin) ;
  
  for(numpoints=0;fgets(buf,LINELEN,stdin);numpoints++)
   { 
     if (numpoints>=MAXCOORDS)
       { puts("Too many input points.") ; return 0 ; }
     if (strncmp(buf,"END",3)==0) return 1 ;
     sscanf(buf,"%d %d",&(coords[numpoints].x),&(coords[numpoints].y)) ;
   }       
  return 0 ;
}  


/* ---------------------------------------------------------------------- */

/* draw a point on screen */

#if VIEW
void put_point(int x,int y,int col)
{int i ;
 vga_setcolor(col) ;
 
 for(i=SCALE*x-2;i<=SCALE*x+2;i++)
  vga_drawline(i,SCALE*y-2,i,SCALE*y+2) ;
}  

/* draw all input points on screen */

void show_input()
{ int i ;

  for(i=0;i<numpoints;i++)
   put_point(coords[i].x,coords[i].y,2) ;
}   
#endif

/* ---------------------------------------------------------------------- */

/* number of distinct x/y coordinates, sorted coordinates */

int numx,numy,sortx[MAXCOORDS],sorty[MAXCOORDS] ;
struct point trcoords[MAXCOORDS] ; /* points in new coordinates */

/* process input, find all distinct x/y coordinates, make the transforms */ 

void process_coords()
{ int i,j,x,y ;
  
  for(i=0;i<numpoints;i++)
   {
    x=coords[i].x ; y=coords[i].y ;
    
    for(j=0;j<numx;j++)
     if (sortx[j]>=x) break ;
     
    if (j==numx) sortx[numx++]=x ;
      else if (sortx[j]!=x)
             { memmove(sortx+j+1,sortx+j,sizeof(int)*(numx-j)) ;
               sortx[j]=x ; numx++ ;
             }
    
    for(j=0;j<numy;j++)
     if (sorty[j]>=y) break ;
     
    if (j==numy) sorty[numy++]=y ;
      else if (sorty[j]!=y)
             { memmove(sorty+j+1,sorty+j,sizeof(int)*(numy-j)) ;
               sorty[j]=y ; numy++ ;
             }
             
   }          

  for(i=0;i<numpoints;i++)
   {
    x=coords[i].x ; y=coords[i].y ;
     
    for(j=0;j<numx;j++)
      if (sortx[j]==x) 
        { trcoords[i].x=j ; break ; }
    
    for(j=0;j<numy;j++)
      if (sorty[j]==y) 
        { trcoords[i].y=j ; break ; }
   }     
}

/* ---------------------------------------------------------------------- */

/* We represent all interesting points in a matrix in transformed 
  coordinates containing the following constants */

#define EMPTY   0
#define STEINER 1
#define BASE    2
#define SOLVED  3


typedef unsigned char nodet ; /* matrix member (node) type */

nodet *nodes ;

/* access node at x,y */
#define N(x,y) (*(nodes+(y)+numy*(x)))

#if VIEW

/* draw all nodes */
void show_nodes()
{ int i,j,c=7 ;

  for(i=0;i<numx;i++)
    for(j=0;j<numy;j++)
     { switch(N(i,j))
        { case EMPTY : c=1 ; break ;
          case STEINER : c=2 ; break ;
          case BASE : c=4 ; break ;
          case SOLVED : c=3 ; break ;
        }
       put_point(sortx[i],sorty[j],c) ;
     }  
}   
#endif

/* ---------------------------------------------------------------------- */


/* A few macros to simplify conditions in the following rules */

#define UPEMPTY(x,y)    ((y)+1>=numy || N(x,(y)+1)==EMPTY)
#define DOWNEMPTY(x,y)  ((y)-1<0 || N(x,(y)-1)==EMPTY)
#define RIGHTEMPTY(x,y) ((x)+1>=numx || N((x)+1,y)==EMPTY)
#define LEFTEMPTY(x,y)  ((x)-1<0 || N((x)-1,y)==EMPTY)

#define UPSTEINER(x,y)   ((y)+1<numy && N(x,(y)+1)==STEINER)
#define DOWNSTEINER(x,y) ((y)>0 && N(x,(y)-1)==STEINER)
#define RIGHTSTEINER(x,y)((x)+1<numx && N((x)+1,y)==STEINER)
#define LEFTSTEINER(x,y) (x>0 && N((x)-1,y)==STEINER)



void change_node(int x,int y) ; /* a node has changed */


/* change state of node x,y according to rules, call itself recursively
   on neighbours to make changes penetrate */
   
void simplify_nodes(int x,int y)
{ int t,u,r,d,l ;

  t=N(x,y) ; 
  
  if (t==STEINER)
    { if (UPEMPTY(x,y) || DOWNEMPTY(x,y))
        if (RIGHTEMPTY(x,y) || LEFTEMPTY(x,y))
          { N(x,y)=EMPTY ;
            change_node(x,y) ;
          }
      return ;    
    }
  if (t==BASE)
    { u=UPEMPTY(x,y) ;
      r=RIGHTEMPTY(x,y) ;
      d=DOWNEMPTY(x,y) ;
      l=LEFTEMPTY(x,y) ;
      if (u+r+d+l!=3) return ;
      
      if (l==0 && LEFTSTEINER(x,y))
        { N(x,y)=SOLVED ;
          N(x-1,y)=BASE ;
          change_node(x,y) ;
          return ;
        }
      if (r==0 && RIGHTSTEINER(x,y))
        { N(x,y)=SOLVED ;
          N(x+1,y)=BASE ;
          change_node(x,y) ;
          return ;
        }
      if (u==0 && UPSTEINER(x,y))
        { N(x,y)=SOLVED ;
          N(x,y+1)=BASE ;
          change_node(x,y) ;
          return ;
        }
      if (d==0 && DOWNSTEINER(x,y))
        { N(x,y)=SOLVED ;
          N(x,y-1)=BASE ;
          change_node(x,y) ;
          return ;
        }
    }
}

void change_node(int x,int y)
{
 if (x>0)      simplify_nodes(x-1,y) ;
 if (x<numx-1) simplify_nodes(x+1,y) ;
 if (y>0)      simplify_nodes(x,y-1) ;
 if (y<numy-1) simplify_nodes(x,y+1) ;
} 
  
/* ---------------------------------------------------------------------- */
   
/* allocate memory for matrix and set the node states */

int make_nodes()
{ int i,j ;

 if ((nodes=malloc(numx*numy*sizeof(nodet)))==NULL)
   { puts("Not enough memory for nodes") ; return 0 ; }
   
 for(i=0;i<numx;i++)
  for(j=0;j<numy;j++)
    N(i,j)=STEINER ;
      
 for(i=0;i<numpoints;i++)
    N(trcoords[i].x,trcoords[i].y)=BASE ;
 
 for(i=0;i<numx;i++)
  for(j=0;j<numy;j++)
   simplify_nodes(i,j) ; 

#if VIEW    
 show_nodes() ;        
#endif
 
 return 1 ;     
}

/* ---------------------------------------------------------------------- */

/* in the space we have numbases points that are to be included and
   numsteiners points that may be included. 
   nnodes is the total number of nodes used for a given try */

int numbases,numsteiners;
struct point *graph ;
int *gd,*uptr,*tptr,*val,*dad  ;  /* graph distances */
int nnodes ;

#define G(i,j) (abs(graph[i].x-graph[j].x)+abs(graph[i].y-graph[j].y))
/* distance between nodes */
   
#define U(t)  (*(uptr+t))
   /* is Steiner's node used ? */
   
#define T(t)  (*(tptr+t))   
   /* mapping */             

int nnused ; /* number of nodes in use */


/* go through the matrix and note the interesting points */

int prepare_graph()
{ int i,j,k ;

  numbases=numsteiners=0 ;

  for(i=0;i<numx;i++)
   for(j=0;j<numy;j++)
    switch(N(i,j))
     { case BASE:    numbases++ ; break ;
       case STEINER: numsteiners++ ; break ;
     }  
  
  nnodes=numbases+numsteiners ;
  
    
  if ((graph=malloc(nnodes*sizeof(struct point)))==NULL)
   { puts("Not enough memory for graph") ; return 0 ; }
   
  if ((uptr=malloc((numsteiners+3*nnodes+3)*sizeof(int)))==NULL)
   { puts("Not enough memory for gd") ; return 0 ; }
   
  tptr=uptr+numsteiners ;
  val=tptr+nnodes+1 ;
  dad=val+nnodes+1 ;
   
  k=0 ;
  
  for(i=0;i<numx;i++)
   for(j=0;j<numy;j++)
    if (N(i,j)==BASE)
     { (graph+k)->x=sortx[i] ; 
       (graph+k++)->y=sorty[j] ;
     }
     
  for(i=0;i<numx;i++)
   for(j=0;j<numy;j++)
    switch(N(i,j))
     { case STEINER: (graph+k)->x=sortx[i] ; 
                     (graph+k++)->y=sorty[j] ;
                     break ;   
     }
  
  return 1 ;      
}

/* ---------------------------------------------------------------------- */

/* Prim algorithm for finding the minimal spanning tree */

#define UNSEEN 1000000

void find_tree()
{ int k,min,t,prio ;

#if DEBUG
  puts("find_tree") ;
#endif 


  for(k=0;k<nnused;k++) { val[k]=-UNSEEN ; dad[k]=-1 ; }
  val[-1]=-(UNSEEN+1) ; /* sentinel */
  min=0 ;
  
  do 
   { k=min ; val[k]=-val[k] ; min=-1 ;
     if (val[k]==UNSEEN) val[k]=0 ;
          
     for(t=0;t<nnused;t++)
      if (val[t]<0) /* nodes not belonging to the tree */ 
        { if ( val[t]<(prio=-G(T(k),T(t))) )
            { val[t]=prio ; dad[t]=k ; }
            
          if (val[t]>val[min]) min=t ;  
        }  
   } while (min>=0) ;
}

/* ---------------------------------------------------------------------- */


/* draw a tree */

#if VIEW 
void draw_tree()
{ int ti,i,d,td,len,x1,x2,y1,y2 ;

  len=0 ;
  vga_setcolor(7) ;

  for(i=0;i<nnused;i++)
     { ti=T(i) ;
     if ((d=dad[i])>-1)
      { td=T(d) ;
        len+=G(ti,td) ;
        x1=graph[ti].x ; y1=graph[ti].y ; x2=graph[td].x ; y2=graph[td].y  ;

        vga_drawline(SCALE*x1,SCALE*y1,SCALE*x1,SCALE*y2) ;
        vga_drawline(SCALE*x1,SCALE*y2,SCALE*x2,SCALE*y2) ;
       }    
    }    
}     
#endif

/* ---------------------------------------------------------------------- */

/* output a TeX code to draw a graph */

void tex_out()
{ int ti,i,d,td,len,x1,x2,y1,y2 ;
  char *c ;
  for(c=test_name;*c && *c!=' ';c++) ;
  *c='\0' ; 
  fprintf(texf,"\\zadani{%s}\n",test_name) ;
  
  for(i=0;i<numpoints;i++)
    fprintf(texf,"\\bod{%d}{%d}\n",coords[i].x,coords[i].y) ;
    
  len=0 ;

  for(i=0;i<nnused;i++)
     { ti=T(i) ;
     if ((d=dad[i])>-1)
      { td=T(d) ;
        len+=G(ti,td) ;
        x1=graph[ti].x ; y1=graph[ti].y ; x2=graph[td].x ; y2=graph[td].y  ;

        if (y1!=y2) fprintf(texf,"\\vusecka{%d}{%d}{%d}\n",x1,y1,y2) ;
        if (x1!=x2) fprintf(texf,"\\husecka{%d}{%d}{%d}\n",y2,x1,x2) ;
       }    
    }    
  
  fprintf(texf,"\\delka{%d}\n",len) ;
}     


/* ---------------------------------------------------------------------- */

/* calculate the cost of the tree */

int measure_tree()
{ int ti,i,d,len ;

  len=0 ;
  for(i=0;i<nnused;i++)
    { ti=T(i) ;
      if ((d=dad[i])>-1) len+=G(ti,T(d)) ; 
    }
    
  return len ;      
}     

/* ---------------------------------------------------------------------- */


int stbmplen ; /* bitmap length for Steiner trees */
int ngen ;
unsigned int *gen ;

/* from a given bitmap, generate a configuration */

void genU(unsigned int *tn)
{ 
  unsigned int mask ;
  int i ;

#if 0
  printf("genU ptr=%p tmp=%u\n",tn,tn[1]) ;
#endif 
  nnused=numbases ;
  
  for(i=0;i<numsteiners;i++) U(i)=0 ;

  for(i=0;i<numsteiners;i++)
   { mask=1 ; mask<<=i%INTBITS ;
     if ((tn[i/INTBITS] & mask)==0) continue ;
     U(i)=1 ;
     T(nnused++)=i+numbases ;
   }
}


/* ---------------------------------------------------------------------- */

/* construct and measure a tree */

int eval_tree()
{ int len ;

  find_tree() ;
  len=measure_tree() ;
  return len ;
}  

/* ---------------------------------------------------------------------- */

/* minimum Steiner tree found for a simplified case, add the segments
   we have not considered */

void finalize()
{ int i,j,k,x,y ;

 for(i=0;i<numx;i++)
  for(j=0;j<numy;j++)
    switch(N(i,j))
     { case SOLVED: N(i,j)=BASE ; 
                    break ;
      
       case STEINER: N(i,j)=EMPTY ; break ;
     }  

 for(i=0;i<numpoints;i++) N(trcoords[i].x,trcoords[i].y)=BASE ;
  
    
 for(i=0;i<numsteiners;i++)
  if (U(i)) 
    { x=graph[i+numbases].x ; y=graph[i+numbases].y ;
    
      for(j=0;j<numx;j++)
        if (x==sortx[j]) break ;
        
      for(k=0;k<numy;k++)
        if (y==sorty[k]) break ;
        
      N(j,k)=STEINER ;
    }

  free(gd) ; free(graph) ; 
  prepare_graph() ;
  
  nnused=nnodes ;  
  
  for(i=0;i<nnodes;i++)      T(i)=i ;
  for(i=0;i<numsteiners;i++) U(i)=1 ;
}

/* ---------------------------------------------------------------------- */

/* alarm handler */

static volatile nosig ;

void catch_alarm(int sig)
{ puts("Signal caught.") ;
  keep_going=0 ;
  if (nosig==0) /* first signal */
    { nosig=1 ; return ; }
  puts("Terminating.") ;
#if VIEW
  vga_setmode(TEXT) ;
#endif  
  signal(sig,SIG_DFL) ; /* next time we terminate */
  raise(sig) ;
}  




#define BITSET(x) gen[(x)/INTBITS]|=((unsigned int)1)<<((x)%INTBITS)
#define BITRES(x) gen[(x)/INTBITS]&=~(((unsigned int)1)<<((x)%INTBITS))
#define ISBIT(x) (gen[(x)/INTBITS] & (((unsigned int)1)<<((x)%INTBITS)))



int a_search()  /* the search algorithm */
{ int i,j,len ;
  int best_len ;
  struct itimerval it ;
  clock_t start,finish ;
 
 
  /* get ready for the alarm and set it */
  keep_going=1 ;
  signal(SIGPROF,catch_alarm) ;
  signal(SIGINT,catch_alarm) ;
  signal(SIGALRM,catch_alarm) ;
  signal(SIGVTALRM,catch_alarm) ;
    
  it.it_interval.tv_sec=0 ;
  it.it_interval.tv_usec=0 ;
  it.it_value.tv_sec=ALLOWEDTIME ;
  it.it_value.tv_usec=0 ;
 
  setitimer(ITIMER_PROF,&it,NULL) ; 
  start=clock() ;                    /* timer start */
 
  stbmplen=1+(numsteiners+INTBITS-1) / INTBITS ; 
  
  if ((gen=malloc(stbmplen*sizeof(unsigned int)))==NULL)
    { puts("Out of memory for gen.") ;
      exit(-1) ; }
      
  /* use no Steiner nodes */
  for(i=0;i<numsteiners;i++) BITRES(i) ;
  
  /* initialize T */
  for(i=0;i<nnodes;i++) T(i)=i ;
  
  
  genU(gen) ;                         /* initial estimate */
  best_len=eval_tree() ;
#if 0  
  printf("numbases=%d numstainers=%d initial_len=%d\n",
          numbases,numsteiners,best_len) ;
#endif  
  if (numsteiners>MAXSTPTS)
      
    /* speedy best-only iterative search */
    for(i=0,j=0;keep_going && i<numsteiners;i++,j++)
     { if (j>=numsteiners) j=0 ;
     
       if (!ISBIT(j))
         { BITSET(j) ;
           genU(gen) ;
           len=eval_tree() ;
	   if (len<best_len) { best_len=len ; i=0 ; continue ; }
	   BITRES(j) ;
         }
     } 
   else
    /* normal best-only iterative search */
    while(keep_going)
      { int bests=-1 ;
        for(j=0;j<numsteiners;j++)
          if (!ISBIT(j))
            { BITSET(j) ;
              genU(gen) ;
	      len=eval_tree() ;
	      if (len<best_len) { best_len=len ; bests=j ; }
	      BITRES(j) ;      
            }
        if (bests==-1) break ;
        BITSET(bests) ; 
      }

  /* final optimization - remove what is not needed */
  for(i=0;keep_going && i<numsteiners;i++)
    if (!ISBIT(i))
      { BITRES(i) ;
        genU(gen) ;
        len=eval_tree() ;
        if (len<=best_len) { best_len=len ; continue ; }
        BITSET(i) ;
      }  


  finalize() ;
   
  find_tree() ;
  len=measure_tree() ;

  finish=clock() ; /* timer stop */
  
  printf("%.2f s used.\n",(float)(finish-start)/CLOCKS_PER_SEC) ;

#if VIEW
  show_nodes() ;        
  draw_tree() ;
  sleep(1) ;
  while (vga_getch()==-1) ;
#endif
  

  printf("final len=%d\n",len) ;

  return 1 ;
}


/* ---------------------------------------------------------------------- */


/* the command line argument gives the name of the TeX file to create */

void main(int argc,char *argv[])
{  
#if VIEW
  vga_init() ;
  vga_setmode(G640x480x16) ;
  vga_clear() ; 
#endif
  
  texf=NULL ;
  if (argc==2)
    if ((texf=fopen(argv[1],"w"))==NULL)
      perror("Cannot open Tex file.") ;

  for(;;)
   { if (!read_input()) break ;
     
     printf("%s numpoints=%d\n",test_name,numpoints) ;

#if VIEW     
     show_input() ;
     while(vga_getch()==-1) ;
#endif     

     process_coords() ;

     if (!make_nodes()) break ; 
     
     if (!prepare_graph()) { free(nodes) ; break ; }
     
     if (!a_search()) { free(nodes) ; free(gd) ; free(graph) ; break ; }
   
     if (texf) tex_out()     ;
     free(nodes) ; free(gd) ; free(graph) ;
#if VIEW
     vga_clear() ;
#endif     
   }   
   
  if (texf) fclose(texf) ; 
#if VIEW
  vga_setmode(TEXT) ;
#endif  
}  
